# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/pod.utils.ipynb (unless otherwise specified).

__all__ = ['PYMEMRI_FOLDER', 'POD_KEYS_FOLDER', 'POD_KEYS_FILENAME', 'POD_KEYS_FULL_FOLDER', 'DEFAULT_POD_KEY_PATH',
           'read_pod_key']

# Cell
from pathlib import Path
from ..data.basic import *

# Cell
PYMEMRI_FOLDER = ".pymemri"
POD_KEYS_FOLDER = "pod_keys"
POD_KEYS_FILENAME = "keys.json"
POD_KEYS_FULL_FOLDER = Path.home() / ".pymemri" / POD_KEYS_FOLDER
DEFAULT_POD_KEY_PATH = POD_KEYS_FULL_FOLDER / POD_KEYS_FILENAME

# Cell
def read_pod_key(key_type, file=DEFAULT_POD_KEY_PATH):
    try:
        json = read_json(file)
    except:
        raise ValueError(f"Trying to read key from {file}, but file or key does not exist") from None
    try:
        key = json[key_type]
        print(f"reading {key_type} from {file}")
        return key
    except:
        raise ValueError(f"{key_type} not specified in {file}") from None


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/itembase.ipynb (unless otherwise specified).

__all__ = ['ALL_EDGES', 'Edge', 'EdgeList', 'T', 'ItemBase', 'Item']

# Cell
# hide
from typing import Optional, Dict, List, Set, Any, Generic, TypeVar, Tuple
from ..imports import *
from datetime import datetime
import uuid

ALL_EDGES = "allEdges"
SOURCE, TARGET, TYPE, EDGE_TYPE, LABEL, SEQUENCE = "_source", "_target", "_type", "_type", "label", "sequence"

# Cell
class Edge():
    """Edges makes a link between two `ItemBase` Items. You won't use this class a lot in practice, as edges are
    abstracted away for normal users. When items are retrieved from the database, the edges are parsed automatically.
    When you add an edge between to items within pymemri, you will often use `ItemBase.add_edge`"""
    def __init__(self, source, target, _type, label=None, sequence=None, created=False, reverse=True):
        self.source   = source
        self.target   = target
        self._type    = _type
        self.label    = label
        self.sequence = sequence
        self.created  = created
        self.reverse  = reverse

    @classmethod
    def from_json(cls, json):
        from .schema import get_constructor
        # we only set the target here
        _type = json[EDGE_TYPE]
        json_target = json[TARGET]
        target_type = json_target["_type"]
        plugin_class = json_target.get("pluginClass", None)
        target_constructor = get_constructor(target_type, plugin_class)
        target = target_constructor.from_json(json_target)
        return cls(source=None, target=target, _type=_type)

    def __repr__(self):
        return f"{self.source} --{self._type}-> {self.target}"

    def update(self, api):
        if self.created:
            api.create_edges([self])

    def __eq__(self, other):
        return self.source is other.source and self.target is other.target \
         and self._type == other._type and self.reverse == other.reverse and self.created == other.created \
         and self.label == other.label

    def traverse(self, start):
        """We can traverse an edge starting from the source to the target or vice versa. In practice we often call
        item.some_edge_type, which calls item.traverse(edgetype), which in turn calls this function."""
        if start == self.source:
            return self.target
        elif start == self.target:
            return self.source
        else:
            raise ValueError

# Cell
T = TypeVar('T')

class EdgeList(list, Generic[T]):
    def __init__(
        self,
        items: List[Edge] = None,
        target_type: type = None,
    ) -> None:
        self.target_type = target_type if target_type is not None else Any
        super().__init__(items if items is not None else list())

    @property
    def targets(self):
        return [edge.target for edge in self]

#     @property
#     def target_type(self):
#         tgt_types = [type(edge.target) for edge in self]
#         if tgt_types.count(tgt_types[0]) == len(tgt_types):
#             return tgt_types[0]
#         return Any

# Cell
# hide
class ItemBase:
    """Provides a base class for all items.
    All items in the schema inherit from this class, and it provides some
    basic functionality for consistency and to enable easier usage."""
    properties: List[str] = list()
    edges: List[str] = list()

    def __init__(self, id: str = None):
        self._date_local_modified = dict()
        self._in_pod: bool = False
        self._new_edges = list()
        self._original_properties = dict()

        self.id: Optional[str] = id

    def __setattr__(self, name, value):
        prev_val = getattr(self, name, None)
        super(ItemBase, self).__setattr__(name, value)

        if name in self.properties and value != prev_val:
            self._date_local_modified[name] = datetime.utcnow()
            if name not in self._original_properties:
                self._original_properties[name] = prev_val

    @property
    def _updated_properties(self):
        return set(self._original_properties.keys())

    def __getattribute__(self, name):
        val = object.__getattribute__(self, name)
        if name in object.__getattribute__(self, "edges"):
            if isinstance(val, Edge):
                return val.traverse(start=self)
            if isinstance(val, list) and len(val) > 0 and isinstance(val[0], Edge):
                return [edge.traverse(start=self) for edge in val]
        return val

    def reset_local_sync_state(self):
        """
        reset_local_sync_state is called when self is created or updated (optionally via bulk) in the PodClient.
        """
        self._original_properties = dict()
        self._date_local_modified = dict()
        self._in_pod = True

    def add_edge(self, name, val):
        """Creates an edge of type name and makes it point to val"""
        if name not in self.__dict__:
            raise NameError(f"object {self} does not have edge with name {name}")

        edge = Edge(self, val, name, created=True)
        existing = object.__getattribute__(self, name)
        if edge not in existing:
            res = existing + [edge]
            self.__setattr__(name, res)
            self._new_edges.append(edge)

    def is_expanded(self):
        """returns whether the node is expanded. An expanded node retrieved nodes that are
        *directly* connected to it
        from the pod, and stored their values via edges in the object."""
        return len(self.get_all_edges()) > 0

    def get_edges(self, name):
        return object.__getattribute__(self, name)

    def get_all_edges(self):
        return [e for attr in self.__dict__.values() if self.attr_is_edge(attr) for e in attr]

    def get_all_edge_names(self):
        return [k for k,v in self.__dict__.items() if self.attr_is_edge(v)]

    def get_property_names(self):
        return [k for k, v in self.__dict__.items() if not type(v) == list]

    @staticmethod
    def attr_is_edge(attr):
        return isinstance(attr, list) and len(attr)>0 and isinstance(attr[0], Edge)

    def update(self, api, edges=True, create_if_not_exists=True, skip_nodes=False):

        if not self.exists(api):
            print(f"creating {self}")
            api.create(self)
        else:
            print(f"updating {self}")
            api.update_item(self)

        if edges:
            api.create_edges(self.get_all_edges())

    def exists(self, api):
        return api.exists(self.id) if self.id else None

    def create_id_if_not_exists(self):
        if self.id is None:
            self.id = uuid.uuid4().hex

    def store(self, client: "PodClient"):
        return client.add_to_store(self)

    def __repr__(self):
        id = self.id
        _type = self.__class__.__name__
        return f"{_type} (#{id})"

    @classmethod
    def from_data(cls, *args, **kwargs):
        edges = dict()
        new_kwargs = dict()
        for k, v in kwargs.items():
            if isinstance(v, ItemBase):
                edge = Edge(None, v, k)
                edges[k] = edge
                new_kwargs[k] = edge
            else:
                new_kwargs[k] = v

        res = cls(*args, **new_kwargs)

        for v in edges.values():
            v.source = res
        return res


# Cell
class Item(ItemBase):
    """Item is the baseclass for all of the data classes."""

    properties = [
        "dateCreated",
        "dateModified",
        "dateServerModified",
        "deleted",
        "externalId",
        "itemDescription",
        "starred",
        "version",
        "id",
        "importJson",
        "pluginClass",
        "isMock",
    ]
    edges = ["changelog", "label", "genericAttribute", "measure", "sharedWith"]

    DATE_PROPERTIES = ['dateCreated', 'dateModified', 'dateServerModified']

    def __init__(
        self,
        dateCreated: datetime = None,
        dateModified: datetime = None,
        dateServerModified: datetime = None,
        deleted: bool = None,
        externalId: str = None,
        itemDescription: str = None,
        starred: bool = None,
        version: str = None,
        id: str = None,
        importJson: str = None,
        pluginClass: str = None,
        isMock: bool = None,
        changelog: list = None,
        label: list = None,
        genericAttribute: list = None,
        measure: list = None,
        sharedWith: list = None
    ):
        super().__init__(id)

        # Properties
        self.dateCreated: Optional[str] = dateCreated
        self.dateModified: Optional[str] = dateModified
        self.dateServerModified: Optional[str] = dateServerModified
        self.deleted: Optional[str] = deleted
        self.externalId: Optional[str] = externalId
        self.itemDescription: Optional[str] = itemDescription
        self.starred: Optional[str] = starred
        self.version: Optional[str] = version
        self.importJson: Optional[str] = importJson
        self.pluginClass: Optional[str] = pluginClass
        self.isMock: Optional[bool] = isMock

        # Edges
        self.changelog: list = changelog if changelog is not None else []
        self.label: list = label if label is not None else []
        self.genericAttribute: list = genericAttribute if genericAttribute is not None else []
        self.measure: list = measure if measure is not None else []
        self.sharedWith: list = sharedWith if sharedWith is not None else []

    @classmethod
    def parse_json(self, cls, json):
        property_kwargs = Item.parse_properties(cls, json)
        edge_kwargs = Item.parse_edges(cls, json)
        return {**property_kwargs, **edge_kwargs}

    @classmethod
    def parse_properties(self, cls, json):
        return {p: json.get(p, None) for p in cls.properties}

    @classmethod
    def parse_edges(self, cls, json):
        all_edges = json.get(ALL_EDGES, None)
        edge_kwargs = dict()
        reverse_edges = [f"~{e}" for e in cls.edges]
        if all_edges is not None:
            for edge_json in all_edges:
                edge = Edge.from_json(edge_json)
                if edge.type in self.edges + reverse_edges:
                    edge_name = self.remove_prefix(edge.type)
                    if edge_name in edge_kwargs:
                        edge_kwargs[edge_name] += [edge]
                    else:
                        edge_kwargs[edge_name] = [edge]
        return edge_kwargs

    @classmethod
    def get_property_types(cls, dates=False) -> Dict[str, type]:
        """
        Infer the property types of all properties in cls.
        Raises ValueError if type anotations for properties are missing in the cls init.
        """
        mro = cls.mro()
        property_types = dict()
        for basecls in reversed(mro[:mro.index(ItemBase)]):
            property_types.update(basecls.__init__.__annotations__)
        property_types = {k: v for k, v in property_types.items() if k in cls.properties}

        if not set(property_types.keys()) == set(cls.properties):
            raise ValueError(f"Item {cls.__name__} has missing property annotations.")

        res = dict()
        for k, v in property_types.items():
            if k[:1] != '_' and k != "private" and not (isinstance(v, list)) \
                            and v is not None and (not (dates == False and k in cls.DATE_PROPERTIES)):
                res[k] = v
        return res

    @classmethod
    def get_edge_types(cls) -> Dict[str, Tuple[str, str]]:
        """
        Infer the types of all edges in cls as tuple (source_type, target_type)
        """
        mro = cls.mro()
        tgt_types = dict()
        for basecls in reversed(mro[:mro.index(ItemBase)]):
            tgt_types.update(basecls.__init__.__annotations__)
        tgt_types = {k: v for k, v in tgt_types.items() if k in cls.edges}

        res = dict()
        for k, v in tgt_types.items():
            if hasattr(v, "__args__") and len(v.__args__):
                v = v.__args__[0]
                if isinstance(v, type):
                    v = v.__name__
            else:
                v = "Any"
            res[k] = (cls.__name__, v)
        return res

    @classmethod
    def remove_prefix(s, prefix="~"):
        return s[1:] if s[0] == "`" else s

    @classmethod
    def from_json(cls, json):
        kwargs = Item.parse_json(cls, json)

        property_types = cls.get_property_types(dates=True)
        for k, v in kwargs.items():
            if v is not None and property_types[k] == datetime:
                # Datetime in pod is in milliseconds
                kwargs[k] = datetime.fromtimestamp(v / 1000.)
        res = cls(**kwargs)
        for e in res.get_all_edges():
            e.source = res
        return res

    def _get_schema_type(self):
        for cls in self.__class__.mro():
            if cls.__name__ != "ItemBase":
                return cls.__name__

    def to_json(self, dates=True):
        res = dict()
        private = getattr(self, "private", [])
        for k, v in self.__dict__.items():
            if k[:1] != '_' and k != "private" and k not in private and not (isinstance(v, list)) \
                            and v is not None and (not (dates == False and k in self.DATE_PROPERTIES)):
                if isinstance(v, datetime):
                    # Save datetimes in milliseconds
                    v = int(v.timestamp() * 1000)
                res[k] = v
        res["type"] = self._get_schema_type()
        return res



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/template.formatter.ipynb (unless otherwise specified).

__all__ = ['TEMPLATE_URL', 'TEMPLATE_BASE_PATH', 'get_remote_url', 'infer_git_info', 'GITLAB_GROUPS',
           'str_to_identifier', 'reponame_to_displayname', 'download_plugin_template', 'get_templates',
           'TemplateFormatter', 'get_template_replace_dict', 'plugin_from_template']

# Cell
# hide
from pathlib import Path
from typing import Dict, Union, List
from fastcore.script import call_parse, Param, store_true
import zipfile
import tempfile
import urllib
from string import Template
import re
import giturlparse
import subprocess

# Cell
# hide
TEMPLATE_URL = "https://gitlab.memri.io/memri/plugin-templates/-/archive/dev/plugin-templates-dev.zip"
TEMPLATE_BASE_PATH = "plugin-templates-dev"

# Cell
# hide

# If the owner of the repository is one of these groups, the CLI requires an additional `user` argument
GITLAB_GROUPS = ["memri", "plugins"]

def get_remote_url(path="."):
    path = Path(path)
    url = subprocess.getoutput([f"cd {path} && git config --get remote.origin.url"])
    if not url:
        raise ValueError(f"You can only run this from a initialized gitlab repository, and '{path}' is not an initialized git repository")
    parsed = giturlparse.parse(url)
    repo_url = parsed.url2https
    if repo_url.endswith(".git"):
        repo_url = repo_url[:-4]
    return repo_url

def infer_git_info(url):
    parsed = giturlparse.parse(url)
    return parsed.owner, parsed.repo

# Cell
# hide
def str_to_identifier(s, lower=True):
    result = re.sub("\W|^(?=\d)", "_", s)
    if lower:
        result = result.lower()
    return result

def reponame_to_displayname(reponame: str) -> str:
    return re.sub("[-_]+", " ", reponame).title()

def download_plugin_template(
    template_name: str, url: str = TEMPLATE_URL, base_path: str = TEMPLATE_BASE_PATH
):
    base_path = str(Path(base_path) / template_name)
    zip_path, _ = urllib.request.urlretrieve(url)
    with zipfile.ZipFile(zip_path, "r") as f:
        result = {name: f.read(name) for name in f.namelist() if base_path in name}
    if len(result) == 0:
        raise ValueError(f"Could not find template: {template_name}")
    result = {k.replace(base_path, "").strip("/"): v.decode("utf-8") for k, v in result.items() if v}
    return result


def get_templates(url: str = TEMPLATE_URL) -> List[str]:
    zip_path, _ = urllib.request.urlretrieve(url)
    with zipfile.ZipFile(zip_path, "r") as f:
        files_split = [name.split("/") for name in f.namelist()]
        result = [fn[1] for fn in files_split if fn[-1] == '' and len(fn) == 3]
    return result

# Cell
# hide
class TemplateFormatter:
    def __init__(
        self,
        template_dict: Dict[str, str],
        replace_dict: Dict[str, str],
        tgt_path: Union[str, Path],
        verbose: bool = False,
    ):
        self.template_dict = template_dict
        self.tgt_path = Path(tgt_path)
        self.replace_dict = replace_dict
        self.verbose = verbose

    def format_content(self, content):
        return Template(content).safe_substitute(self.replace_dict)

    def format_path(self, path):
        new_path = Template(path).safe_substitute(self.replace_dict)
        return self.tgt_path / new_path

    def format_file(self, filename, content):
        new_path = self.format_path(filename)
        new_content = self.format_content(content)
        new_path.parent.mkdir(exist_ok=True, parents=True)
        if self.verbose:
            print(f"Formatting {filename} -> {new_path}")
        with open(new_path, "w", encoding="utf-8") as f:
            f.write(new_content)

    def get_files(self):
        return [self.format_path(filename) for filename in self.template_dict.keys()]

    def format(self):
        for filename, content in self.template_dict.items():
            self.format_file(filename, content)

    def print_filetree(self):
        previous_prefix = None
        res = "Created the following files"

        for path in sorted([x.relative_to(self.tgt_path) for x in self.get_files()],
                           key=lambda item: 100 * str(item).count("/")):
            n_slashes = str(path).count("/")
            new_prefix = path.parent
            if previous_prefix != new_prefix and str(new_prefix) != ".":
                res = f"{res}\n├── {new_prefix}"
            if n_slashes == 0:
                res = f"{res}\n├── {path}"
            elif n_slashes == 1:
                res = f"{res}\n│   ├── {path.name}"

            previous_prefix=new_prefix

        print(res.strip() + "\n")

# Cell
# hide
def get_template_replace_dict(
    repo_url=None, user=None, plugin_name=None, package_name=None, description=None
):
    if repo_url is None:
        repo_url = get_remote_url()

    try:
        repo_owner, repo_name = infer_git_info(repo_url)
    except ValueError:
        url_inf, owner_inf, name_inf = None, None, None
        print("Could not infer git information from current directory, no initialized repository found.")

    if repo_url is None:
        repo_url = url_inf

    if user is None:
        if repo_owner in GITLAB_GROUPS:
            user = None
        else:
            user = repo_owner

    if plugin_name is None:
        if repo_name is None:
            print("henk")
            plugin_name = None
        else:
            plugin_name = reponame_to_displayname(repo_name)

    if package_name is None:
        if repo_name is None:
            package_name = None
        else:
            package_name = str_to_identifier(repo_name)

    return {
        "user": user,
        "package_name": package_name,
        "plugin_name": plugin_name,
        "repo_name": repo_name,
        "repo_url": repo_url,
        "description": str(description),
    }

# Cell
@call_parse
def plugin_from_template(
    list_templates: Param("List available plugin templates", store_true) = False,
    user: Param("Your Gitlab username", str) = None,
    repo_url: Param("The url of your empty Gitlab plugin repository", str) = None,
    plugin_name: Param("Display name of your plugin", str) = None,
    template_name: Param(
        "Name of the template, use `list_templates` to see all available options"
    ) = "basic",
    package_name: Param("Name of your plugin python package", str) = None,
    description: Param("Description of your plugin", str) = None,
    target_dir: Param("Directory to output the formatted template", str) = ".",
):
    """
    CLI that downloads and formats a plugin template according to the arguments, and local git repository.

    Args:
        list_templates (Param, optional): If True, only list available templates. Defaults to False.
        user (Param, optional): Your GitLab username. Defaults to None.
        repo_url (Param, optional): The url of your gitlab plugin repository. Defaults to None.
        plugin_name (Param, optional): The name of your plugin. Defaults to None.
        template_name (Param, optional): The name of the template used. To list all options, see `list_templates`.
            Defaults to "basic".
        package_name (Param, optional): The name of the python package of your plugin. Inferred if left blank. Defaults to None.
        description (Param, optional): An optional plugin description. Defaults to None.
        target_dir (Param, optional): Directory where the plugin template is generated. Defaults to ".".
    """
    if list_templates:
        print("Available templates:")
        for template in get_templates():
            print(template)
        return

    template = download_plugin_template(template_name)

    tgt_path = Path(target_dir)
    replace_dict = get_template_replace_dict(
        repo_url=repo_url,
        user=user,
        plugin_name=plugin_name,
        package_name=package_name,
        description=description,
    )

    formatter = TemplateFormatter(template, replace_dict, tgt_path)
    formatter.format()
    formatter.print_filetree()


    print(f"Created `{replace_dict['plugin_name']}` using the {template_name} template.")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plugin.pluginbase.ipynb (unless otherwise specified).

__all__ = ['POD_FULL_ADDRESS_ENV', 'POD_TARGET_ITEM_ENV', 'POD_OWNER_KEY_ENV', 'POD_AUTH_JSON_ENV',
           'POD_PLUGIN_DNS_ENV', 'PluginBase', 'PluginError', 'ExamplePlugin', 'write_run_info', 'get_plugin_cls',
           'run_plugin_from_run_id', 'store_keys', 'parse_metadata', 'parse_config', 'create_run_expanded',
           'run_plugin', 'simulate_run_plugin_from_frontend']

# Cell
from ..data.schema import *
from ..pod.client import *
from ..imports import *
from .states import *
from ..pod.utils import *
from .listeners import get_abort_plugin_listener

from os import environ
from abc import ABCMeta
import abc
import json
import importlib
import string
import time
from enum import Enum
from fastscript import *
import os
from .schema import Account, PluginRun
from ..data.basic import *
from ..pod.client import Dog, PodClient, DEFAULT_POD_ADDRESS
import warnings
from ..data.basic import write_json
from .authenticators.credentials import PLUGIN_DIR
from fastcore.script import call_parse, Param
import os
import traceback

# Cell
POD_FULL_ADDRESS_ENV        = 'POD_FULL_ADDRESS'
POD_TARGET_ITEM_ENV         = 'POD_TARGET_ITEM'
POD_OWNER_KEY_ENV           = 'POD_OWNER'
POD_AUTH_JSON_ENV           = 'POD_AUTH_JSON'
POD_PLUGIN_DNS_ENV          = 'PLUGIN_DNS'

# Cell
class PluginBase(metaclass=ABCMeta):
    """Base class for plugins"""

    schema_classes = []

    def __init__(self, pluginRun=None, client=None, **kwargs):
        super().__init__()
        if pluginRun is None:
            warnings.warn(
                "Plugin needs a pluginRun as kwarg, running without will only work in development.",
                RuntimeWarning)
        self.pluginRun = pluginRun

        if client is None:
            raise ValueError("Plugins need a `client: PodClient` as kwarg to run.")
        self.client = client
        self._status_listeners = []
        self._config_dict = kwargs

    def set_run_status(self, status):
        # TODO sync before setting status (requires pod_client.sync())
        if self.pluginRun and self.client:
            self.pluginRun.status = status
            self.client.update_item(self.pluginRun)

    def set_progress(self, progress):
        if self.pluginRun and self.client:
            self.pluginRun.progress = progress
            self.client.update_item(self.pluginRun)

    def setup(self):
        if self.client and self.pluginRun:
            status_abort_listener = get_abort_plugin_listener(self.client, self.pluginRun.id)
            self._status_listeners.append(status_abort_listener)

    def teardown(self):
        for listener in self._status_listeners:
            listener.stop()

    def _run(self):
        self.setup()
        self.run()
        self.teardown()

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError()

    def add_to_schema(self):
        """
        Add all schema classes required by the plugin to self.client here.
        """
        if len(self.schema_classes):
            self.client.add_to_schema(*self.schema_classes)

    @classmethod
    def get_schema_properties(cls):
        schema = []
        for item in cls.schema_classes:
            item_schema = PodClient._property_dicts_from_type(item)
            schema.extend(item_schema)
        return schema

    @classmethod
    def get_schema_edges(cls):
        schema = []
        for item in cls.schema_classes:
            edge_types = item.get_edge_types()
            edge_schema = [
                {"type": "ItemEdgeSchema",
                 "edgeName": k,
                 "sourceType": v[0],
                 "targetType": v[1]}
                for k, v in edge_types.items()
            ]
            schema.extend(edge_schema)
        return schema

    @classmethod
    def get_schema(cls, include_edges: bool = True):
        schema = cls.get_schema_properties()
        if include_edges:
            edges = cls.get_schema_edges()
            schema.extend(edges)
        return schema

# Cell
# hide
class PluginError(Exception):
    """Generic class for plugin errors. This error is raised when a plugin raises an unexpected exception."""
    pass

# Cell
# hide
class ExamplePlugin(PluginBase):
    schema_classes = [Dog, Message]

    def __init__(self, dog_name: str = "Bob", **kwargs):
        super().__init__(**kwargs)
        self.dog_name = dog_name

    def run(self):
        print("Started plugin run...")
        dog = Dog(self.dog_name, 10)
        self.client.create(dog)
        print("Run success!")

# Cell
# hide
def write_run_info(plugin, id_):
    try:
        if plugin is None:
            raise ValueError("Empty container")
        run_path = PLUGIN_DIR / plugin / "current_run.json"
        run_path.parent.mkdir(parents=True, exist_ok=True)
        print(f"writing run info to {run_path}")
        write_json({"id": id_}, run_path)
    except Exception as e:
        print(f"""failed to write run info to {run_path}\n{e}""")

# Cell
# export
def get_plugin_cls(plugin_module, plugin_name):
    try:
        module = importlib.import_module(plugin_module)
        plugin_cls = getattr(module, plugin_name)
        return plugin_cls
    except (ImportError, AttributeError):
        raise ImportError(f"Unknown plugin: {plugin_module}.{plugin_name}")

def run_plugin_from_run_id(run_id, client, **kwargs):
    """
    Runs a plugin from run_id, initialized with **kwargs.

    Args:
        client (PodClient): client containing PluginRun
        run_id (int): id of the PluginRun
    """

    run = client.get(run_id)
    write_run_info(run.pluginModule.split(".")[0] if run.pluginModule is not None else run.containerImage, run.id)

    plugin_cls = get_plugin_cls(run.pluginModule, run.pluginName)
    plugin = plugin_cls(pluginRun=run, client=client, **kwargs)
    plugin.add_to_schema()

    plugin.set_run_status(RUN_STARTED)
    plugin._run()
    plugin.pluginRun = plugin.client.get(run_id)
    plugin.set_run_status(RUN_COMPLETED)

    return plugin

# Cell
# hide
def _parse_env():
    env = os.environ
    print("Reading `run_plugin()` parameters from environment variables")
    try:
        pod_full_address = env.get(POD_FULL_ADDRESS_ENV, DEFAULT_POD_ADDRESS)
        plugin_run_json  = json.loads(str(env[POD_TARGET_ITEM_ENV]))
        print(plugin_run_json)
        plugin_run_id    = plugin_run_json["id"]
        owner_key        = env.get(POD_OWNER_KEY_ENV)
        pod_auth_json    = json.loads(str(env.get(POD_AUTH_JSON_ENV)))
        return pod_full_address, plugin_run_id, pod_auth_json, owner_key
    except KeyError as e:
        raise Exception('Missing parameter: {}'.format(e)) from None

# Cell
# hide
@call_parse
def store_keys(path:Param("path to store the keys", str)=DEFAULT_POD_KEY_PATH,
               database_key:Param("Database key of the pod", str)=None,
               owner_key:Param("Owner key of the pod", str)=None,
               replace: Param("Replace existing stored keys", str)=True):

    if not replace:
        try:
            read_pod_key("database_key")
            read_pod_key("owner_key")
            print("Existing stored keys found, exiting without generating new keys.")
            return
        except ValueError:
            pass

    if database_key is None: database_key = PodClient.generate_random_key()
    if owner_key is None: owner_key = PodClient.generate_random_key()

    obj = {"database_key": database_key,
           "owner_key": owner_key}
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        timestr = time.strftime("%Y%m%d-%H%M%S")
        path.rename(POD_KEYS_FULL_FOLDER / f"keys-{timestr}.json")
    write_json(obj, path)

# Cell
# hide
def parse_metadata(fn, remove_container=False):
    metadata = read_json(fn)
    for k in ["pluginModule", "pluginName"]:
        if k not in metadata:
            raise ValueError(f"Missing metadata: {k}")

    run_vars = {k: v for k, v in metadata.items() if k in PluginRun.properties}
    run = PluginRun.from_json(run_vars)
    if remove_container:
        run.containerImage = "none"

    if "account" in metadata:
        account = Account.from_json(metadata["account"])
        run.add_edge("account", account)
    return run


def parse_config(run_config, config_file=None, remove_container=False):
    """
    Parse the configuration of the plugin. A configuration is a dict that is passed to the plugin init as kwargs.
    If configuration file is defined, the run_config is ignored.
    """
    if config_file is not None:
        config = read_json(config_file)
    elif isinstance(run_config, str) and len(run_config):
        config = json.loads(run_config)
    else:
        config = dict()

    if not isinstance(config, dict):
        raise ValueError(f"Incorrect plugin config format, expected a dict, got a {type(config)}")
    return config


def create_run_expanded(client, run):
    client.create(run)
    accounts = run.account
    if accounts:
        account=accounts[0]
        client.create(account)
        client.create_edge(run.get_edges("account")[0])

# Cell
@call_parse
def run_plugin(
    pod_full_address: Param("The pod full address", str) = DEFAULT_POD_ADDRESS,
    plugin_run_id: Param("Run id of the plugin to be executed", str) = None,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
    read_args_from_env: Param("Read the args from the environment", bool) = False,
    metadata: Param("metadata file for the PluginRun", str) = None,
    config_file: Param(
        "A plugin configuration, overwrites the configuration of the PluginRun", str
    ) = None,
):

    if read_args_from_env:
        pod_full_address, plugin_run_id, pod_auth_json, owner_key = _parse_env()
        database_key = None
    else:
        if database_key is None:
            database_key = read_pod_key("database_key")
        if owner_key is None:
            owner_key = read_pod_key("owner_key")
        pod_auth_json = None
    if POD_PLUGIN_DNS_ENV in os.environ:
        print(f"Plugin accesible via {os.environ.get(POD_PLUGIN_DNS_ENV)}:8080")

    client = PodClient(
        url=pod_full_address,
        database_key=database_key,
        owner_key=owner_key,
        auth_json=pod_auth_json,
    )
    print(f"pod_full_address={pod_full_address}\nowner_key={owner_key}\n")

    if metadata is not None:
        run = parse_metadata(metadata, remove_container=True)
        create_run_expanded(client, run)
        plugin_run_id = run.id
    else:
        run = client.get(plugin_run_id)
    plugin_config = parse_config(run.config, config_file)

    try:
        run_plugin_from_run_id(
            plugin_run_id, client, **plugin_config
        )
    except Exception as e:
        run = client.get(plugin_run_id)
        run.status = RUN_FAILED
        client.update_item(run)
        print(traceback.format_exc(), flush=True)
        raise PluginError("The plugin quit unexpectedly.") from None

# Cell
@call_parse
def simulate_run_plugin_from_frontend(
    pod_full_address: Param("The pod full address", str) = DEFAULT_POD_ADDRESS,
    database_key: Param("Database key of the pod", str) = None,
    owner_key: Param("Owner key of the pod", str) = None,
    container: Param("Pod container to run frod", str) = None,
    plugin_path: Param("Plugin path", str) = None,
    metadata: Param("metadata file for the PluginRun", str) = None,
    config_file: Param(
        "A plugin configuration, overwrites the configuration of the PluginRun", str
    ) = None,
    account_id: Param("Account id to be used inside the plugin", str) = None,
):
    if database_key is None:
        database_key = read_pod_key("database_key")
    if owner_key is None:
        owner_key = read_pod_key("owner_key")
    params = [pod_full_address, database_key, owner_key]
    if None in params:
        raise ValueError(f"Missing Pod credentials")

    client = PodClient(url=pod_full_address, database_key=database_key, owner_key=owner_key)
    print(f"pod_full_address={pod_full_address}\nowner_key={owner_key}\n")

    if metadata is not None:
        run = parse_metadata(metadata)
        create_run_expanded(client, run)
    else:
        if container is None:
            container = plugin_path.split(".", 1)[0]
        print(f"Inferred '{container}' as plugin container name")
        plugin_module, plugin_name = plugin_path.rsplit(".", 1)
        run = PluginRun(container, plugin_module, plugin_name)

        if account_id is not None:
            account = client.get(account_id)
            run.add_edge("account", account)
            print(f"Using existing {account}")

        client.create(run)

    print(
        f"Created pluginrun with id {run.id} on {pod_full_address}"
    )

    plugin_dir = run.containerImage
    write_run_info(plugin_dir, run.id)

    print(f"*Check the pod log/console for debug output.*")
    return run


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plugin.stateful.ipynb (unless otherwise specified).

__all__ = ['PersistentState', 'RUN_IDLE', 'RUN_INITIALIZED', 'RUN_USER_ACTION_NEEDED', 'RUN_USER_ACTION_COMPLETED',
           'RUN_STARTED', 'RUN_FAILED', 'RUN_COMPLETED', 'StatefulPlugin', 'MyStatefulPlugin']

# Cell
from ..data.schema import Item
from .pluginbase import PluginBase
import logging

# Cell
class PersistentState(Item):
    """ Persistent state variables saved for plugin such as views, accounts, the last state to resume from etc. """

    properties = Item.properties + ["pluginId", "state"]
    edges = Item.edges + ["account", "view"]

    def __init__(self, pluginName=None, state=None, account=None, view=None, **kwargs):
        super().__init__(**kwargs)
        self.pluginName = pluginName
        self.state = state
        self.account = account if account is not None else []
        self.view = view if view is not None else []

    def get_state(self):
        return self.state

    def set_state(self, client, state_str):
        self.state = state_str
        client.update_item(self)

    def get_account(self):
        if len(self.account) == 0:
            return None
        else:
            return self.account[0]

    def set_account(self, client, account):
        if len(self.account) == 0:
            if not account.id:
                client.create(account)
            self.add_edge('account', account)
            self.update(client)
        else:
            existing_account = self.account[0]
            for prop in account.properties:
                value = getattr(account, prop, None)
                if value and hasattr(existing_account, prop):
                    setattr(existing_account, prop, value)
            existing_account.update(client)

    def get_view_by_name(self, view_name):
        for cvu in self.view:
            if cvu.name == view_name:
                return cvu

    def set_views(self, client, views=None):
        for view in views:
            client.create(view)
            self.add_edge('view', view)
        self.update(client)
        return True

# Cell
RUN_IDLE = 'idle'                           #1
RUN_INITIALIZED = 'initilized'              #2
RUN_USER_ACTION_NEEDED = 'userActionNeeded' # 2-3
RUN_USER_ACTION_COMPLETED = 'ready'         # 2-3
RUN_STARTED = 'start'                       #3
RUN_FAILED = 'error'                        # 3-4
RUN_COMPLETED = 'done'                      #4

logging.basicConfig(format='%(asctime)s [%(levelname)s] - %(message)s')

# Cell
class StatefulPlugin(PluginBase):
    """ Provides state/view setter and getter functions to plugin runtime """

    properties = PluginBase.properties + ["runId", "persistenceId"]
    edges = PluginBase.edges

    def __init__(self, runId=None, persistenceId=None, **kwargs):
        super().__init__(**kwargs)
        self.runId = runId
        self.persistenceId = persistenceId

    def persist(self, client, pluginName, views=None, account=None):
        persistence = PersistentState(pluginName=pluginName)
        client.create(persistence)
        self.persistenceId = persistence.id
        if views:
            persistence.set_views(client, views)
        if account:
            persistence.set_account(account)

    def get_state(self, client, pluginName=None):
        if self.persistenceId:
            return client.get(self.persistenceId)
        elif pluginName:
            result = client.search({'type': 'PersistentState', 'pluginName': pluginName})
            if len(result) > 0:
                self.persistenceId = result[0].id
                return self.get_state(client)

    def set_account(self, client, account):
        state = self.get_state(client)
        state.set_account(account)

    def set_state_str(self, client, state_str):
        state = self.get_state(client)
        state.set_state(client, state_str)

    def initialized(self, client):
        logging.warning("PLUGIN run is initialized")
        self.set_run_vars(client, {'state':RUN_INITIALIZED})

    def started(self, client):
        logging.warning("PLUGIN run is started")
        self.set_run_vars(client, {'state':RUN_STARTED})

    def failed(self, client, error):
        logging.error(f"PLUGIN run is failed: {error}")
        print("Exception while running plugin:", error)
        self.set_run_vars(client, {'state':RUN_FAILED, 'error': str(error)})

    def completed(self, client):
        logging.warning("PLUGIN run is completed")
        self.set_run_vars(client, {'state':RUN_COMPLETED})

    def complete_action(self, client):
        self.set_run_vars(client, {'state': RUN_USER_ACTION_COMPLETED})

    def action_required(self, client):
        self.set_run_vars(client, {'state': RUN_USER_ACTION_NEEDED})

    def is_action_required(self, client):
        return self.get_run_state(client) == RUN_USER_ACTION_NEEDED

    def is_action_completed(self, client):
        return self.get_run_state(client) == RUN_USER_ACTION_COMPLETED

    def is_completed(self, client):
        return self.get_run_state(client) == RUN_COMPLETED

    def is_failed(self, client):
        return self.get_run_state(client) == RUN_FAILED

    def is_daemon(self, client):
        run = self.get_run(client, expanded=False)
        return run.interval and run.interval > 0

    def get_run(self, client, expanded=False):
        return client.get(self.runId, expanded=expanded)

    def get_run_state(self, client):
        start_plugin = self.get_run(client)
        return start_plugin.status

    def set_run_vars(self, client, vars):
        start_plugin = client.get(self.runId, expanded=False)
        for k,v in vars.items():
            if hasattr(start_plugin, k):
                setattr(start_plugin, k, v)
        client.update_item(start_plugin)

    def get_run_view(self, client):
        run = self.get_run(client, expanded=True)
        if run:
            for view in run.view:
                return view

    def set_run_view(self, client, view_name):
        state = self.get_state(client)
        view = state.get_view_by_name(view_name)

        if view:
            attached_CVU_edge = self.get_run_view(client) # index error here if there is no already bound CVU
            if attached_CVU_edge:
                logging.warning(f"Plugin Run already has a view. Updating with {view_name}")
                attached_CVU_edge.target = view  # update CVU
                attached_CVU_edge.update(client) # having doubts if this really updates the existing edge
            else:
                logging.warning(f"Plugin Run does not have a view. Creating {view_name}")
                run = self.get_run(client)
                run.add_edge('view', view)
                run.update(client)
            return True
        return False

    def add_to_schema(self, client):
        assert client.add_to_schema(PersistentState("", ""))

# Cell
# hide
from ..data.schema import Person

class MyStatefulPlugin(StatefulPlugin):

    def __init__(self, runId=None, **kwargs):
        super().__init__(runId=runId, **kwargs)

    def run(self, client):
        # plugin's magic happens here

        # manipulate run state
        self.set_run_vars({'state': 'Running'})

        # create items in POD
        imported_person = Person(firstName="Hari", lastName="Seldon")
        client.create(imported_person)

        # set persistent state
        self.set_state_str("continue_from:5021")


    def add_to_schema(self, client):
        print("Adding schema")
        super().add_to_schema(client)
        # add plugin-specific schemas here
        client.add_to_schema(Person(firstName="", lastName=""))
        pass


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/basic.ipynb (unless otherwise specified).

__all__ = ['read_file', 'read_json', 'write_json', 'unzip', 'PYI_HOME', 'PYI_TESTDATA', 'HOME_DIR', 'MODEL_DIR',
           'MEMRI_S3']

# Cell
from ..imports import *
from urllib.request import urlretrieve
import requests
from tqdm import tqdm
import zipfile

# Cell
Path.ls = lambda x: list(x.iterdir())
PYI_HOME = Path.cwd().parent
PYI_TESTDATA = PYI_HOME / "test" / "data"
HOME_DIR = Path.home()
MODEL_DIR = HOME_DIR / ".memri" / "models"
MEMRI_S3 = "https://memri.s3-eu-west-1.amazonaws.com"
MODEL_DIR.mkdir(parents=True, exist_ok=True)


def read_file(path):
    return open(path, "r").read()

def read_json(path):
    with open(path) as json_file:
        return json.load(json_file)

def write_json(obj, fname, indent=4):
    with open(fname, 'w') as file_out:
        json.dump(obj , file_out, indent=indent)

def unzip(f, dest):
    with zipfile.ZipFile(str(f)) as zf:
        zf.extractall(str(dest))


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plugin.states.ipynb (unless otherwise specified).

__all__ = ['RUN_IDLE', 'RUN_INITIALIZED', 'RUN_USER_ACTION_NEEDED', 'RUN_USER_ACTION_COMPLETED', 'RUN_STARTED',
           'RUN_FAILED', 'RUN_COMPLETED', 'RUN_ABORTED', 'RUN_AUTHENTICATED']

# Cell

# run states
RUN_IDLE = 'idle'
RUN_INITIALIZED = 'initialized'
RUN_USER_ACTION_NEEDED = 'userActionNeeded'
RUN_USER_ACTION_COMPLETED = 'ready'
RUN_STARTED = 'started'
RUN_FAILED = 'error'
RUN_COMPLETED = 'done'
RUN_ABORTED = 'aborted'
RUN_AUTHENTICATED = 'authenticated'


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"read_file": "basic.ipynb",
         "read_json": "basic.ipynb",
         "write_json": "basic.ipynb",
         "unzip": "basic.ipynb",
         "Path.ls": "basic.ipynb",
         "PYI_HOME": "basic.ipynb",
         "PYI_TESTDATA": "basic.ipynb",
         "HOME_DIR": "basic.ipynb",
         "MODEL_DIR": "basic.ipynb",
         "MEMRI_S3": "basic.ipynb",
         "CVU_BASE_PATH": "cvu.utils.ipynb",
         "get_default_cvu": "cvu.utils.ipynb",
         "list_default_cvus": "cvu.utils.ipynb",
         "filter_rows": "data.dataset.ipynb",
         "Dataset": "data.dataset.ipynb",
         "MEMRI_PATH": "data.loader.ipynb",
         "MEMRI_GITLAB_BASE_URL": "data.loader.ipynb",
         "ACCESS_TOKEN_PATH": "data.loader.ipynb",
         "GITLAB_API_BASE_URL": "data.loader.ipynb",
         "DEFAULT_PLUGIN_MODEL_PACKAGE_NAME": "data.loader.ipynb",
         "DEFAULT_PYTORCH_MODEL_NAME": "data.loader.ipynb",
         "DEFAULT_HUGGINFACE_CONFIG_NAME": "data.loader.ipynb",
         "DEFAULT_PACKAGE_VERSION": "data.loader.ipynb",
         "TIME_FORMAT_GITLAB": "data.loader.ipynb",
         "PROJET_ID_PATTERN": "data.loader.ipynb",
         "find_git_repo": "data.loader.ipynb",
         "get_registry_api_key": "data.loader.ipynb",
         "upload_in_chunks": "data.loader.ipynb",
         "IterableToFileAdapter": "data.loader.ipynb",
         "write_file_to_package_registry": "data.loader.ipynb",
         "project_id_from_name": "data.loader.ipynb",
         "get_project_id_from_project_path_unsafe": "data.loader.ipynb",
         "write_huggingface_model_to_package_registry": "data.loader.ipynb",
         "write_model_to_package_registry": "data.loader.ipynb",
         "download_package_file": "data.loader.ipynb",
         "download_huggingface_model_for_project": "data.loader.ipynb",
         "load_huggingface_model_for_project": "data.loader.ipynb",
         "DEFAULT_ENCODING": "data.photo.ipynb",
         "show_images": "data.photo.ipynb",
         "Photo": "data.photo.ipynb",
         "Query": "exporters.exporters.ipynb",
         "ALL_EDGES": "itembase.ipynb",
         "Edge": "itembase.ipynb",
         "EdgeList": "itembase.ipynb",
         "T": "itembase.ipynb",
         "ItemBase": "itembase.ipynb",
         "Item": "itembase.ipynb",
         "PLUGIN_DIR": "plugin.authenticators.credentials.ipynb",
         "read_username_password": "plugin.authenticators.credentials.ipynb",
         "OAuthAuthenticator": "plugin.authenticators.oauth.ipynb",
         "StatusListener": "plugin.listeners.ipynb",
         "force_exit_callback": "plugin.listeners.ipynb",
         "get_abort_plugin_listener": "plugin.listeners.ipynb",
         "POD_FULL_ADDRESS_ENV": "plugin.pluginbase.ipynb",
         "POD_TARGET_ITEM_ENV": "plugin.pluginbase.ipynb",
         "POD_OWNER_KEY_ENV": "plugin.pluginbase.ipynb",
         "POD_AUTH_JSON_ENV": "plugin.pluginbase.ipynb",
         "POD_PLUGIN_DNS_ENV": "plugin.pluginbase.ipynb",
         "PluginBase": "plugin.pluginbase.ipynb",
         "PluginError": "plugin.pluginbase.ipynb",
         "ExamplePlugin": "plugin.pluginbase.ipynb",
         "write_run_info": "plugin.pluginbase.ipynb",
         "get_plugin_cls": "plugin.pluginbase.ipynb",
         "run_plugin_from_run_id": "plugin.pluginbase.ipynb",
         "store_keys": "plugin.pluginbase.ipynb",
         "parse_metadata": "plugin.pluginbase.ipynb",
         "parse_config": "plugin.pluginbase.ipynb",
         "create_run_expanded": "plugin.pluginbase.ipynb",
         "run_plugin": "plugin.pluginbase.ipynb",
         "simulate_run_plugin_from_frontend": "plugin.pluginbase.ipynb",
         "RUN_IDLE": "plugin.states.ipynb",
         "RUN_INITIALIZED": "plugin.states.ipynb",
         "RUN_USER_ACTION_NEEDED": "plugin.states.ipynb",
         "RUN_USER_ACTION_COMPLETED": "plugin.states.ipynb",
         "RUN_STARTED": "plugin.states.ipynb",
         "RUN_FAILED": "plugin.states.ipynb",
         "RUN_COMPLETED": "plugin.states.ipynb",
         "RUN_ABORTED": "plugin.states.ipynb",
         "RUN_AUTHENTICATED": "plugin.states.ipynb",
         "PasswordAuthenticator": "plugins.authenticators.password.ipynb",
         "set_account_credentials": "plugins.authenticators.password.ipynb",
         "simulate_enter_credentials": "plugins.authenticators.password.ipynb",
         "DEFAULT_POD_ADDRESS": "pod.api.ipynb",
         "POD_VERSION": "pod.api.ipynb",
         "PodError": "pod.api.ipynb",
         "PodAPI": "pod.api.ipynb",
         "PodClient": "pod.client.ipynb",
         "Dog": "pod.client.ipynb",
         "Priority": "pod.db.ipynb",
         "DB": "pod.db.ipynb",
         "PYMEMRI_FOLDER": "pod.utils.ipynb",
         "POD_KEYS_FOLDER": "pod.utils.ipynb",
         "POD_KEYS_FILENAME": "pod.utils.ipynb",
         "POD_KEYS_FULL_FOLDER": "pod.utils.ipynb",
         "DEFAULT_POD_KEY_PATH": "pod.utils.ipynb",
         "read_pod_key": "pod.utils.ipynb",
         "get_params": "template.config.ipynb",
         "identifier_to_displayname": "template.config.ipynb",
         "get_param_config": "template.config.ipynb",
         "ALLOWED_TYPES": "template.config.ipynb",
         "create_config": "template.config.ipynb",
         "create_plugin_config": "template.config.ipynb",
         "TEMPLATE_URL": "template.formatter.ipynb",
         "TEMPLATE_BASE_PATH": "template.formatter.ipynb",
         "get_remote_url": "template.formatter.ipynb",
         "infer_git_info": "template.formatter.ipynb",
         "GITLAB_GROUPS": "template.formatter.ipynb",
         "str_to_identifier": "template.formatter.ipynb",
         "reponame_to_displayname": "template.formatter.ipynb",
         "download_plugin_template": "template.formatter.ipynb",
         "get_templates": "template.formatter.ipynb",
         "TemplateFormatter": "template.formatter.ipynb",
         "get_template_replace_dict": "template.formatter.ipynb",
         "plugin_from_template": "template.formatter.ipynb",
         "get_ci_variables": "test_utils.ipynb"}

modules = ["data/basic.py",
           "cvu/utils.py",
           "data/dataset.py",
           "data/loader.py",
           "data/photo.py",
           "exporters/exporters.py",
           "data/itembase.py",
           "plugin/authenticators/credentials.py",
           "plugin/authenticators/oauth.py",
           "plugin/listeners.py",
           "plugin/pluginbase.py",
           "plugin/states.py",
           "plugin/authenticators/password.py",
           "pod/api.py",
           "pod/client.py",
           "pod/db.py",
           "pod/utils.py",
           "template/config.py",
           "template/formatter.py",
           "test_utils.py"]

doc_url = "http://memri.docs.memri.io/pymemri/"

git_url = "https://gitlab.memri.io/memri/pymemri/tree/prod/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/test_utils.ipynb (unless otherwise specified).

__all__ = ['get_ci_variables']

# Cell
#hide
import os

def get_ci_variables(*varnames):
    """
    Gets ci variables from environment, can be set on gitlab in settings -> ci/cd -> variables
    """
    return {varname: os.environ.get(varname) for varname in varnames}


# For backwards compatibility with plugins, import from central schema here
# TODO remove this file and update plugins.

__all__ = ["Account", "PluginRun"]

from ..data.schema import Account, PluginRun


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/pod.client.ipynb (unless otherwise specified).

__all__ = ['PodClient', 'Dog']

# Cell
from ..data.basic import *
from ..data.schema import *
from ..data.itembase import Edge, ItemBase, Item
from ..imports import *
from hashlib import sha256
from .db import DB, Priority
from .utils import *
from ..plugin.schema import *
from ..test_utils import get_ci_variables
from .api import PodAPI, PodError, DEFAULT_POD_ADDRESS, POD_VERSION

from typing import List, Union
import uuid
import urllib
from datetime import datetime
import warnings

from threading import Thread

# Cell
class PodClient:
    # Mapping from python type to schema type
    # TODO move to data.schema once schema is refactored
    TYPE_TO_SCHEMA = {
        bool: "Bool",
        str: "Text",
        int: "Integer",
        float: "Real",
        datetime: "DateTime",
    }

    def __init__(
        self,
        url=DEFAULT_POD_ADDRESS,
        version=POD_VERSION,
        database_key=None,
        owner_key=None,
        auth_json=None,
        register_base_schema=True,
        verbose=False,
        default_priority=Priority.local,
    ):
        self.verbose = verbose
        self.database_key = (
            database_key if database_key is not None else self.generate_random_key()
        )
        self.owner_key = (
            owner_key if owner_key is not None else self.generate_random_key()
        )
        self.api = PodAPI(
            database_key=self.database_key,
            owner_key=self.owner_key,
            url=url,
            version=version,
            auth_json=auth_json,
            verbose=verbose,
        )

        self.default_priority = Priority(default_priority)
        self.api.test_connection()
        self.local_db = DB()
        self.registered_classes = dict()
        self.register_base_schemas()

    @classmethod
    def from_local_keys(cls, path=DEFAULT_POD_KEY_PATH, **kwargs):
        return cls(
            database_key=read_pod_key("database_key"),
            owner_key=read_pod_key("owner_key"),
            **kwargs,
        )

    @staticmethod
    def generate_random_key():
        return "".join([str(random.randint(0, 9)) for i in range(64)])

    def register_base_schemas(self):
        assert self.add_to_schema(PluginRun, CVUStoredDefinition, Account, Photo)

    def add_to_store(self, item: Item, priority: Priority = None) -> Item:
        item.create_id_if_not_exists()
        priority = priority if priority is not None else self.default_priority
        return self.local_db.merge(item, priority)

    def reset_local_db(self):
        self.local_db = DB()

    def get_create_dict(self, item):
        properties = item.to_json()
        properties = {k: v for k, v in properties.items() if v != []}
        return properties

    def create(self, item):
        self.add_to_store(item)
        create_dict = self.get_create_dict(item)
        try:
            result = self.api.create_item(create_dict)
            item.id = result
            item.reset_local_sync_state()
            if getattr(item, "requires_client_ref", False):
                item._client = self
            return True
        except Exception as e:
            print(e)
            return False

    def create_photo(self, photo, asyncFlag=True):
        file = photo.file[0]

        # create the photo
        items_edges_success = self.bulk_action(
            create_items=[photo, file], create_edges=photo.get_edges("file")
        )
        if not items_edges_success:
            raise ValueError("Could not create file or photo item")

        return self._upload_image(photo.data, asyncFlag=asyncFlag)

        return self._upload_image(photo.data)

    @classmethod
    def _property_dicts_from_instance(cls, item):
        create_items = []
        attributes = item.to_json()
        for k, v in attributes.items():
            if type(v) not in cls.TYPE_TO_SCHEMA:
                raise ValueError(f"Could not add property {k} with type {type(v)}")
            value_type = cls.TYPE_TO_SCHEMA[type(v)]

            create_items.append(
                {
                    "type": "ItemPropertySchema",
                    "itemType": attributes["type"],
                    "propertyName": k,
                    "valueType": value_type,
                }
            )
        return create_items

    @classmethod
    def _property_dicts_from_type(cls, item):
        create_items = []
        for property, p_type in item.get_property_types().items():
            p_type = cls.TYPE_TO_SCHEMA[p_type]
            create_items.append(
                {
                    "type": "ItemPropertySchema",
                    "itemType": item.__name__,
                    "propertyName": property,
                    "valueType": p_type,
                }
            )
        return create_items

    def add_to_schema(self, *items: List[Union[object, type]]):
        create_items = []

        for item in items:
            if isinstance(item, type):
                property_dicts = self._property_dicts_from_type(item)
            else:
                property_dicts = self._property_dicts_from_instance(item)
                item = type(item)
            create_items.extend(property_dicts)
            self.registered_classes[item.__name__] = item

        try:
            self.api.bulk(create_items=create_items)
            return True
        except Exception as e:
            print(e)
            return False

    def _upload_image(self, img, asyncFlag=True, callback=None):
        if isinstance(img, np.ndarray):
            return self.upload_file(img.tobytes(), asyncFlag=asyncFlag, callback=callback)
        elif isinstance(img, bytes):
            return self.upload_file(img, asyncFlag=asyncFlag, callback=callback)
        else:
            raise ValueError(f"Unknown image data type {type(img)}")

    def upload_file(self, file, asyncFlag=True, callback=None):
        if asyncFlag:
            return self.upload_file_async(file, callback=callback)
        try:
            self.api.upload_file(file)
            return True
        except PodError as e:
            # 409 = CONFLICT, file already exists
            if e.status == 409:
                return True
            return False

    def upload_file_async(self, file, callback=None):
        def thread_fn(file, callback):
            result = self.upload_file(file, asyncFlag=False)
            if callback:
                callback(result)
        Thread(target=thread_fn, args=(file, callback)).start()
        return True

    def get_file(self, sha):
        return self.api.get_file(sha)

    def get_photo(self, id, size=640):
        photo = self.get(id)
        self._load_photo_data(photo, size=size)
        return photo

    def _load_photo_data(self, photo, size=None):
        if len(photo.file) > 0 and photo.data is None:
            file = self.get_file(photo.file[0].sha256)
            if file is None:
                print(
                    f"Could not load data of {photo} attached file item does not have data in pod"
                )
                return
            photo.data = file
        else:
            print(f"could not load data of {photo}, no file attached")

    def create_if_external_id_not_exists(self, item):
        if not self.external_id_exists(item):
            self.create(item)

    def external_id_exists(self, item):
        if item.externalId is None:
            return False
        existing = self.search({"externalId": item.externalId})
        return len(existing) > 0

    def create_edges(self, edges):
        return self.bulk_action(create_edges=edges)

    def delete_items(self, items):
        return self.bulk_action(delete_items=items)

    def delete_all(self):
        items = self.get_all_items()
        self.delete_items(items)

    @staticmethod
    def gather_batch(items, start_idx, start_size=0, max_size=5000000):
        idx = start_idx
        total_size = start_size
        batch_items = []
        for i, x in enumerate(items):
            if i < idx:
                continue
            elif len(str(x)) > max_size:
                idx = i + 1
                print("Could not add item: Item exceeds max item size")
            elif total_size + len(str(x)) < max_size:
                batch_items.append(x)
                total_size += len(str(x))
                idx = i + 1
            else:
                break
        return batch_items, idx, total_size

    def bulk_action(
        self,
        create_items=None,
        update_items=None,
        create_edges=None,
        delete_items=None,
        partial_update=True,
        priority=None,
    ):
        priority = Priority(priority) if priority else None

        all_items = []
        if create_items:
            all_items += create_items
        if update_items:
            all_items += update_items

        # we need to add to local_db to not lose reference.
        if create_items is not None:
            for c in create_items:
                if getattr(c, "requires_client_ref", False):
                    c._client = self
                if not self.local_db.contains(c):
                    self.add_to_store(c, priority=priority)

        create_items = (
            [self.get_create_dict(i) for i in create_items]
            if create_items is not None
            else []
        )
        update_items = (
            [
                self.get_update_dict(i, partial_update=partial_update)
                for i in update_items
            ]
            if update_items is not None
            else []
        )
        create_edges = (
            [self.get_create_edge_dict(i) for i in create_edges]
            if create_edges is not None
            else []
        )
        # Note: skip delete_items without id, as items that are not in pod cannot be deleted
        delete_items = (
            [item.id for item in delete_items if item.id is not None]
            if delete_items is not None
            else []
        )

        n_total = len(create_items + update_items + create_edges + delete_items)
        n = 0

        i_ci, i_ui, i_ce, i_di = 0, 0, 0, 0
        while not (
            i_ci == len(create_items)
            and i_ui == len(update_items)
            and i_ce == len(create_edges)
            and i_di == len(delete_items)
        ):
            batch_size = 0
            create_items_batch, i_ci, batch_size = self.gather_batch(
                create_items, i_ci, start_size=batch_size
            )
            update_items_batch, i_ui, batch_size = self.gather_batch(
                update_items, i_ui, start_size=batch_size
            )
            delete_items_batch, i_di, batch_size = self.gather_batch(
                delete_items, i_di, start_size=batch_size
            )
            if i_ci == len(create_items):
                create_edges_batch, i_ce, batch_size = self.gather_batch(
                    create_edges, i_ce, start_size=batch_size
                )
            else:
                create_edges_batch = []
            n_batch = len(
                create_items_batch
                + update_items_batch
                + create_edges_batch
                + delete_items_batch
            )
            n += n_batch
            print(f"BULK: Writing {n}/{n_total} items/edges")

            try:
                result = self.api.bulk(
                    create_items_batch,
                    update_items_batch,
                    create_edges_batch,
                    delete_items_batch,
                )
            except PodError as e:
                print(e)
                print("could not complete bulk action, aborting")
                return False
        print(f"Completed Bulk action, written {n} items/edges")

        for item in all_items:
            item.reset_local_sync_state()
        return True

    def get_create_edge_dict(self, edge):
        return {
            "_source": edge.source.id,
            "_target": edge.target.id,
            "_name": edge._type,
        }

    def create_edge(self, edge):
        edge_dict = self.get_create_edge_dict(edge)

        try:
            self.api.create_edge(edge_dict)
            return True
        except PodError as e:
            print(e)
            return False

    def get(self, id, expanded=True, include_deleted=False):
        if not expanded:
            res = self._get_item_with_properties(id)
        else:
            res = self._get_item_expanded(id)
        if res is None:
            raise ValueError(f"Item with id {id} does not exist")
        elif res.deleted and not include_deleted:
            print(f"Item with id {id} has been deleted")
            return
        return res

    def get_all_items(self):
        raise NotImplementedError()

    def filter_deleted(self, items):
        return [i for i in items if not i.deleted == True]

    def _get_item_expanded(self, id):
        item = self._get_item_with_properties(id)
        edges = self.get_edges(id)
        for e in edges:
            item.add_edge(e["name"], e["item"])
        return item

    def get_edges(self, id):
        try:
            result = self.api.get_edges(id)
            for d in result:
                edge_item = self.item_from_json(d["item"])
                edge_item.reset_local_sync_state()
                d["item"] = edge_item
            return result
        except PodError as e:
            print(e)
            return

    def _get_item_with_properties(self, id):
        try:
            result = self.api.get_item(str(id))
            if not len(result):
                return
            item = self.item_from_json(result[0])
            item.reset_local_sync_state()
            return item
        except PodError as e:
            print(e)
            return

    def get_update_dict(self, item, partial_update=True):
        properties = item.to_json(dates=False)
        properties.pop("type", None)
        properties.pop("deleted", None)
        properties = {
            k: v
            for k, v in properties.items()
            if k == "id" or k in item._updated_properties
        }
        return properties

    def update_item(self, item, partial_update=True):
        data = self.get_update_dict(item, partial_update=partial_update)
        try:
            self.api.update_item(data)
            item.reset_local_sync_state()
            return True
        except PodError as e:
            print(e)
            return False

    def exists(self, id):
        try:
            result = self.api.get_item(str(id))
            if isinstance(result, list) and len(result) > 0:
                return True
            return False
        except PodError as e:
            print(e)
            return False

    def search_paginate(
        self,
        fields_data,
        limit=50,
        include_edges=True,
        add_to_local_db: bool = True,
        priority=None,
    ):
        priority = Priority(priority) if priority else None

        if "ids" in fields_data:
            raise NotImplementedError(
                "Searching by multiple IDs is not implemented for paginated search."
            )

        extra_fields = {"[[edges]]": {}} if include_edges else {}
        query = {**fields_data, **extra_fields}

        try:
            for page in self.api.search_paginate(query, limit):
                result = [
                    self._item_from_search(
                        item, add_to_local_db=add_to_local_db, priority=priority
                    )
                    for item in page
                ]
                yield self.filter_deleted(result)
        except PodError as e:
            print(e)

    def search(
        self,
        fields_data,
        include_edges: bool = True,
        add_to_local_db: bool = True,
        priority=None,
    ):
        priority = Priority(priority) if priority else None

        extra_fields = {"[[edges]]": {}} if include_edges else {}
        query = {**fields_data, **extra_fields}

        # Special key "ids" for searching a list of ids.
        # Requires /bulk search instead of /search.
        if "ids" in query:
            ids = query.pop("ids")
            bulk_query = [{"id": uid, **query} for uid in ids]
            try:
                result = self.api.bulk(search=bulk_query)["search"]
            except PodError as e:
                print(e)

            result = [item for sublist in result for item in sublist]
        else:
            try:
                result = self.api.search(query)
            except PodError as e:
                print(e)

        result = [
            self._item_from_search(
                item, add_to_local_db=add_to_local_db, priority=priority
            )
            for item in result
        ]
        return self.filter_deleted(result)

    def _item_from_search(
        self, item_json: dict, add_to_local_db: bool = True, priority=None
    ):
        # search returns different fields w.r.t. edges compared to `get` api,
        # different method to keep `self.get` clean.
        item = self.item_from_json(
            item_json, add_to_local_db=add_to_local_db, priority=priority
        )
        item.reset_local_sync_state()

        for edge_json in item_json.get("[[edges]]", []):
            edge_name = edge_json["_edge"]
            try:
                edge_item = self.item_from_json(
                    edge_json["_item"],
                    add_to_local_db=add_to_local_db,
                    priority=priority,
                )
                edge_item.reset_local_sync_state()
                item.add_edge(edge_name, edge_item)
            except Exception as e:
                print(f"Could not attach edge {edge_json['_item']} to {item}")
                print(e)
                continue
        return item

    def search_last_added(self, type=None, with_prop=None, with_val=None):
        query = {"_limit": 1, "_sortOrder": "Desc"}
        if type is not None:
            query["type"] = type
        if with_prop is not None:
            query[f"{with_prop}=="] = with_val
        return self.search(query)[0]

    def item_from_json(
        self,
        json: dict,
        add_to_local_db: bool = True,
        priority=None,
    ) -> Item:
        priority = Priority(priority) if priority else None

        plugin_class = json.get("pluginClass", None)
        plugin_package = json.get("pluginPackage", None)
        item_class = get_constructor(
            json["type"],
            plugin_class,
            plugin_package=plugin_package,
            extra=self.registered_classes,
        )

        new_item = item_class.from_json(json)
        if add_to_local_db:
            new_item = self.add_to_store(new_item, priority=priority)
        if getattr(new_item, "requires_client_ref", False):
            new_item._client = self
        return new_item

    def get_properties(self, expanded):
        properties = copy(expanded)
        if ALL_EDGES in properties:
            del properties[ALL_EDGES]
        return properties

    def send_email(self, to, subject="", body=""):
        try:
            self.api.send_email(to, subject, body)
            print(f"succesfully sent email to {to}")
            return True
        except PodError as e:
            print(e)
            return False

    def sync(self, priority: str = Priority.newest):
        priority = Priority(priority) if priority else None
        all_items = list(self.local_db.nodes.values())
        all_ids = set(self.local_db.nodes.keys())

        create_items = []
        update_items = []
        create_edges = []

        # Add items from sync store
        for item in all_items:
            if isinstance(item, File) or isinstance(item, Photo):
                continue
            if item._in_pod:
                update_items.append(item)
            else:
                create_items.append(item)

            # Add all edges where src and tgt exist
            for edge in item._new_edges:
                if edge.target.id in all_ids:
                    create_edges.append(edge)
                else:
                    warnings.warn(
                        f"Could not sync `{edge._type}` for {item}: edge target missing in sync."
                    )

        update_ids = [item.id for item in update_items]
        existing_items = self.search(
            {"ids": update_ids}, add_to_local_db=True, priority=priority
        )

        return self.bulk_action(
            create_items=create_items,
            update_items=update_items,
            create_edges=create_edges,
            priority=priority,
        )

    def get_dataset(self, name):
        datasets = self.search({"type": "Dataset", "name": name})
        if len(datasets) == 0:
            raise PodError(f"No datasets found with name {name}")
        elif len(datasets) > 1:
            warnings.warn(f"Multiple datasets found with name {name}. Using the newest dataset.")
        return datasets[-1]

# Cell
class Dog(Item):
    properties = Item.properties + ["name", "age", "bites", "weight"]

    def __init__(self, name: str=None, age: int=None, bites: bool=False, weight: float=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.age = age
        self.bites = bites
        self.weight = weight


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plugin.authenticators.credentials.ipynb (unless otherwise specified).

__all__ = ['PLUGIN_DIR', 'read_username_password']

# Cell
from ...data.basic import read_json, HOME_DIR
PLUGIN_DIR = HOME_DIR / ".pymemri" / "plugins"

# Cell
def read_username_password(plugin: str):
    credentials_path = PLUGIN_DIR / plugin / "credentials.json"
    credentials = read_json(credentials_path)
    username, password = credentials["username"], credentials["password"]
    return username,password


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plugin.listeners.ipynb (unless otherwise specified).

__all__ = ['StatusListener', 'force_exit_callback', 'get_abort_plugin_listener']

# Cell

import time
import os
import signal
import threading
from threading import Thread
from .states import RUN_COMPLETED, RUN_FAILED

# Cell

class StatusListener:
    def __init__(self, client, run_id, status, callback, interval=5, verbose=False):
        self.client = client
        self.run_id = run_id
        self.status = status
        self.callback = callback
        self.interval = interval
        self.verbose = verbose
        self.running = True

    def stop(self):
        if self.verbose:
            print("Stopping listener...", flush=True)
        self.running = False

    def run(self):
        if self.verbose:
            print(f"Listening for status='{self.status}' on Item {self.run_id}", flush=True)

        while self.running and threading.main_thread().is_alive():
            time.sleep(self.interval)
            try:
                run = self.client.get(self.run_id)
                if self.verbose:
                    print("run status:", run.status, flush=True)
                if run.status == self.status:
                    self.callback()
            except Exception as e:
                print(f"Could not get run in status listener: {e}")

# Cell
def force_exit_callback():
    print("Status aborted, killing plugin...", flush=True)
    pid = os.getpid()
    os.kill(pid, signal.SIGINT)

def get_abort_plugin_listener(client, run_id, **kwargs):
    listener = StatusListener(
        client=client,
        run_id=run_id,
        status="aborted",
        callback=force_exit_callback,
        **kwargs
    )
    thread = Thread(
        target=listener.run,
    )
    thread.start()

    return listener


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/pod.api.ipynb (unless otherwise specified).

__all__ = ['DEFAULT_POD_ADDRESS', 'POD_VERSION', 'PodError', 'PodAPI']

# Cell
import os
from typing import Any, Dict, List, Generator, Deque
import requests
import urllib
from hashlib import sha256
from collections import deque
import json

# Cell
DEFAULT_POD_ADDRESS = os.environ.get("POD_ADDRESS") or "http://localhost:3030"
POD_VERSION = "v4"

# Cell
class PodError(Exception):
    def __init__(self, status=None, message=None, **kwargs) -> None:
        super().__init__(status, message, **kwargs)
        self.status = status
        self.message = message

    def __str__(self) -> str:
        return " ".join([str(a) for a in self.args if a])

# Cell
class PodAPI:
    def __init__(
        self,
        database_key: str,
        owner_key: str,
        url: str = DEFAULT_POD_ADDRESS,
        version: str = POD_VERSION,
        auth_json: dict = None,
        verbose: bool = True,
    ) -> None:
        self.verbose = verbose
        self.database_key = database_key
        self.owner_key = owner_key
        self.version = version
        self._url = url
        self.base_url = f"{url}/{version}/{self.owner_key}"
        self.auth_json = self._create_auth(auth_json)

    def _create_auth(self, auth_json: dict = None) -> dict:
        if auth_json is not None:
            return {"type": "PluginAuth", **auth_json}
        else:
            return {"type": "ClientAuth", "databaseKey": self.database_key}

    def test_connection(self) -> bool:
        try:
            res = requests.get(self._url)
            if self.verbose:
                print("Succesfully connected to pod")
            return True
        except requests.exceptions.RequestException as e:
            print("Could no connect to backend")
            return False

    @property
    def pod_version(self) -> dict:
        response = requests.get(f"{self._url}/version")
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response.json()

    def post(self, endpoint: str, payload: Any) -> Any:
        body = {"auth": self.auth_json, "payload": payload}
        response = requests.post(f"{self.base_url}/{endpoint}", json=body)
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response

    def get_item(self, uid: str) -> dict:
        return self.post("get_item", uid).json()

    def create_item(self, item: dict) -> str:
        return self.post("create_item", item).json()

    def update_item(self, item: dict) -> list:
        return self.post("update_item", item).json()

    def get_edges(
        self, uid: str, direction: str = "Outgoing", expand_items: bool = True
    ) -> List[dict]:
        payload = {"item": uid, "direction": direction, "expandItems": expand_items}
        return self.post("get_edges", payload).json()

    def create_edge(self, edge: dict) -> str:
        return self.post("create_edge", edge).json()

    def delete_item(self, uid) -> list:
        return self.post("delete_item", uid).json()

    def search(self, query: dict) -> List[dict]:
        return self.post("search", query).json()

    def search_paginate(self, query: dict, limit: int = 32, even_page_size=True) -> Generator:
        """
        The Pod returns uneven page sizes when paginating, which can be an issue for some applications.
        `search_paginate` wraps the pagination, and always returns pages of size `limit` by storing overflow items
        in a queue.
        """
        paginator = self._paginate(query, limit)

        if not even_page_size:
            yield from paginator

        remaining: Deque[Item] = deque()
        while True:
            if len(remaining) >= limit:
                yield [remaining.popleft() for _ in range(limit)]
            if len(remaining) < limit:
                try:
                    remaining.extend(next(paginator))
                except StopIteration:
                    break

        while len(remaining):
            yield [remaining.popleft() for _ in range(min(limit, len(remaining)))]


    def _paginate(self, query: dict, limit: int = 32) -> Generator:
        if (
            "_limit" in query
            or "dateServerModified" in query
            or "dateServerModified>=" in query
            or "dateServerModified<" in query
        ):
            raise ValueError("Cannot paginate query that contains a date or limit.")
        if "_sortOrder" in query:
            raise NotImplementedError("Only 'Asc' order is supported.")

        query = {**query, "_limit": limit}
        response = self.search(query)
        if not len(response):
            return
            yield

        next_date = 0
        while True:
            query["dateServerModified>="] = next_date
            response = self.search(query)
            if not len(response):
                break

            next_date = response[-1]["dateServerModified"] + 1
            yield response

    def bulk(
        self,
        create_items: List[dict] = None,
        update_items: List[dict] = None,
        create_edges: List[dict] = None,
        delete_items: List[str] = None,
        search: List[dict] = None,
    ) -> Dict[str, Any]:

        payload = {
            "createItems": create_items,
            "updateItems": update_items,
            "createEdges": create_edges,
            "deleteItems": delete_items,
            "search": search,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self.post("bulk", payload).json()

    def upload_file(self, file: bytes) -> Any:
        if self.auth_json.get("type") == "PluginAuth":
            # alternative file upload for plugins, with different authentication
            return self.upload_file_b(file)

        sha = sha256(file).hexdigest()
        result = requests.post(
            f"{self.base_url}/upload_file/{self.database_key}/{sha}", data=file
        )
        if result.status_code != 200:
            raise PodError(result.status_code, result.text)

        return result

    def upload_file_b(self, file: bytes) -> Any:
        sha = sha256(file).hexdigest()
        auth = urllib.parse.quote(json.dumps(self.auth_json))
        result = requests.post(f"{self.base_url}/upload_file_b/{auth}/{sha}", data=file)
        if result.status_code != 200:
            raise PodError(result.status_code, result.text)

        return result

    def get_file(self, sha: str) -> bytes:
        return self.post("get_file", {"sha256": sha}).content

    def send_email(self, to: str, subject: str = "", body: str = "") -> Any:
        payload = {"to": to, "subject": subject, "body": body}
        return self.post("send_email", payload)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/data.loader.ipynb (unless otherwise specified).

__all__ = ['MEMRI_PATH', 'MEMRI_GITLAB_BASE_URL', 'ACCESS_TOKEN_PATH', 'GITLAB_API_BASE_URL',
           'DEFAULT_PLUGIN_MODEL_PACKAGE_NAME', 'DEFAULT_PYTORCH_MODEL_NAME', 'DEFAULT_HUGGINFACE_CONFIG_NAME',
           'DEFAULT_PACKAGE_VERSION', 'TIME_FORMAT_GITLAB', 'PROJET_ID_PATTERN', 'find_git_repo',
           'get_registry_api_key', 'upload_in_chunks', 'IterableToFileAdapter', 'write_file_to_package_registry',
           'project_id_from_name', 'get_project_id_from_project_path_unsafe',
           'write_huggingface_model_to_package_registry', 'write_model_to_package_registry', 'download_package_file',
           'download_huggingface_model_for_project', 'load_huggingface_model_for_project']

# Cell
from fastprogress.fastprogress import progress_bar
from pathlib import Path
import requests
import os, sys
from getpass import getpass
from datetime import datetime
from git import Repo
import re

# Cell
MEMRI_PATH = Path.home() / ".memri"
MEMRI_GITLAB_BASE_URL = "https://gitlab.memri.io"
ACCESS_TOKEN_PATH = Path.home() / ".memri/access_token/access_token.txt"
GITLAB_API_BASE_URL = "https://gitlab.memri.io/api/v4"
DEFAULT_PLUGIN_MODEL_PACKAGE_NAME = "plugin-model-package"
DEFAULT_PYTORCH_MODEL_NAME = "pytorch_model.bin"
DEFAULT_HUGGINFACE_CONFIG_NAME = "config.json"
DEFAULT_PACKAGE_VERSION = "0.0.1"

TIME_FORMAT_GITLAB = '%Y-%m-%dT%H:%M:%S.%fZ'
PROJET_ID_PATTERN = '(?<=<span class="gl-button-text">Project ID: )159(?=</span>)'

# Cell
def find_git_repo():
    path = "."
    for i in range(10):
        try:
            repo = Repo(f"{path + ('.' * i)}/")
        except:
            pass
        else:
            break
    if i == 9:
        raise ValueError(f"could not fine git repo in {os.path.abspath('')}")

    repo_name = repo.remotes.origin.url.split('.git')[0].split('/')[-1]
    return repo_name

# Cell
def get_registry_api_key():
    ACCESS_TOKEN_PATH.parent.mkdir(parents=True, exist_ok=True)
    if ACCESS_TOKEN_PATH.is_file():
        with open(ACCESS_TOKEN_PATH, "r") as f:
            return f.read()
    else:
        print(f"""
        The first time you are uploading a model you need to create an access_token
        at https://gitlab.memri.io/-/profile/personal_access_tokens?name=Model+Access+token&scopes=api
        Click at the blue button with 'Create personal access token'"
        """)
        access_token = getpass("Then copy your personal access token from 'Your new personal access token', and paste here: ")
        with open(ACCESS_TOKEN_PATH, "w") as f:
            f.write(access_token)
        return access_token

# Cell
class upload_in_chunks(object):
    def __init__(self, filename, chunksize=1 << 13):
        self.filename = filename
        self.chunksize = chunksize
        self.totalsize = os.path.getsize(filename)
        self.readsofar = 0

    def __iter__(self):
        n = 1000
        pb = progress_bar(range(n))
        pb_iter = iter(pb)
        i = 1
        delta = 1 / n

        with open(self.filename, 'rb') as file:
            while True:
                data = file.read(self.chunksize)
                if not data:
                    sys.stderr.write("\n")
                    break
                self.readsofar += len(data)
                percent = self.readsofar * 1e2 / self.totalsize
                while (percent / 100) > i * delta:
                    next(pb_iter, None)
                    i += 1
                yield data
        pb.update_bar(n)

    def __len__(self):
        return self.totalsize

class IterableToFileAdapter(object):
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.length = len(iterable)

    def read(self, size=-1): # TBD: add buffer for `len(data) > size` case
        return next(self.iterator, b'')

    def __len__(self):
        return self.length

# Cell
def write_file_to_package_registry(project_id, file_path, api_key, version=DEFAULT_PACKAGE_VERSION):
    file_path = Path(file_path)
    file_name = file_path.name

    url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/packages/generic/{DEFAULT_PLUGIN_MODEL_PACKAGE_NAME}/{version}/{file_name}"
    print(f"uploading {file_path}")
    it = upload_in_chunks(file_path, 10)
    res = requests.put(url=url, data=IterableToFileAdapter(it),
                     headers={"PRIVATE-TOKEN": api_key})

    if res.status_code not in [200, 201]:
        print(f"Failed to upload {file_path}: {res.content}")
    else:
        print(f"Succesfully uploaded {file_path}")

# Cell
def project_id_from_name(project_name, api_key, job_token=None):
    if api_key:
        headers = {"PRIVATE-TOKEN": api_key}
    else:
        headers = {"JOB-TOKEN": job_token}
    res = requests.get(f"{GITLAB_API_BASE_URL}/projects",
                       headers=headers,
                       params={
                           "owned": True,
                           "search": project_name
                       })
    res =  [x.get("id") for x in res.json()]
    if len(res) == 0:
        raise ValueError(f"No plugin found with name {project_name}")
    else:
        return res[0]

# Cell
def get_project_id_from_project_path_unsafe(project_path):
    try:
        res = requests.get(f"{MEMRI_GITLAB_BASE_URL}/{project_path}")
        html = str(res.content)
        match = re.search(PROJET_ID_PATTERN, html)
        return match.group()
    except Exception:
        raise ValueError(f"Could not find project with name {project_path}")

# Cell
def write_huggingface_model_to_package_registry(project_name, model):
    import torch
    api_key = get_registry_api_key()
    project_id = project_id_from_name(project_name, api_key)
    local_save_dir = Path("/tmp")
    torch.save(model.state_dict(), local_save_dir / DEFAULT_PYTORCH_MODEL_NAME)
    model.config.to_json_file(local_save_dir / DEFAULT_HUGGINFACE_CONFIG_NAME)

    for f in [DEFAULT_HUGGINFACE_CONFIG_NAME, DEFAULT_PYTORCH_MODEL_NAME]:
        file_path = local_save_dir / f
        print(f"writing {f} to package registry of {project_name} with project id {project_id}")
        write_file_to_package_registry(project_id, file_path, api_key)

# Cell
def write_model_to_package_registry(model, project_name=None):
    project_name = project_name if project_name is not None else find_git_repo()
    if type(model).__module__.startswith("transformers"):
        import transformers
        import torch
    if isinstance(model, transformers.PreTrainedModel):
        write_huggingface_model_to_package_registry(project_name, model)
    else:
        raise ValueError(f"Model type not supported: {type(model)}")

# Cell
def download_package_file(filename, project_path=None, out_dir=None, package_name=DEFAULT_PLUGIN_MODEL_PACKAGE_NAME,
                          package_version=DEFAULT_PACKAGE_VERSION, download_if_exists=False):
#     if project_name is None:
#         try:
#             project_name = find_git_repo()
#         except Exception as e:
#             raise ValueError("no project name provided, but could also not find a git repo to infer project name") from None
    project_name = str(project_path).split("/")[-1]
    out_dir = out_dir if out_dir is not None else MEMRI_PATH / "projects" / project_name
    out_dir.mkdir(parents=True, exist_ok=True)
#     if os.environ.get("CI", False):
#         api_key = get_registry_api_key()
#         job_token = None
#     else:
#         api_key=None
#         job_token = os.environ.get("CI_JOB_TOKEN")

    project_id = get_project_id_from_project_path_unsafe(project_path)

#     project_id = project_id_from_name(project_name, api_key, job_token)
    file_path = out_dir / filename

    if file_path.exists() and not download_if_exists:
        print(f"{file_path} already exists, and `download_if_exists`==False, using cached version")
        return out_dir

    print(f"downloading {filename} from project {project_path}, package {package_name}")

    res = requests.get(
        url=f"{GITLAB_API_BASE_URL}/projects/{project_id}/packages/generic/{package_name}/{package_version}/{filename}"
    )
    res.raise_for_status()
    with open(out_dir / filename, "wb") as f:
        print(f"writing {filename} to {out_dir}")
        f.write(res.content)
    return file_path

# Cell
def download_huggingface_model_for_project(project_path=None, files=None, download_if_exists=False):
    if files is None:
        files = ["config.json", "pytorch_model.bin"]
    for f in files:
        out_file_path = download_package_file(f, project_path=project_path)
    return out_file_path.parent

# Cell
def load_huggingface_model_for_project(project_path=None, files=None, download_if_exists=False):
    out_dir = download_huggingface_model_for_project(project_path, files, download_if_exists)
    from transformers import AutoModelForSequenceClassification
    model = AutoModelForSequenceClassification.from_pretrained("distilroberta-base", num_labels=10)
    return model